# -*- coding: utf-8 -*-
from translations_server.lib import db


_get_translation = db.SelectOne('''
    SELECT translation_text
    FROM translations
    JOIN languages USING (language_id)
    LEFT JOIN countries USING (country_id)
    WHERE
        language_code = %s AND (country_code IS NULL OR country_code = %s) AND
        translation_key = %s AND
        ((%s = '%%' AND plural IS NULL) OR plural = %s)
    ORDER BY country_code ASC -- existing country comes first
    LIMIT 1
''')


def get_translation(lang, country, key, plural):
    """
    :type lang: str
    :type country: str
    :type key: str
    :type plural: int or None
    :rtype: str or None
    """
    return _get_translation(
        lang, country or None, key, "%" if plural is None else 'x',
        plural)


# The following SQL statements are used by tests.
_insert_translation = db.Manipulation('''
INSERT INTO translations
    (translation_key, language_id, country_id, plural, translation_text)
VALUES (%s, %s, %s, %s, %s)
''')

_insert_language = db.SelectOne('''
    INSERT INTO languages (language_id, language_code, name)
    VALUES (DEFAULT, %s, %s)
    RETURNING language_id
''')

_insert_country = db.SelectOne('''
    INSERT INTO countries (country_id, country_code, name)
    VALUES (DEFAULT, %s, %s)
    RETURNING country_id
''')



# -*- coding: utf-8 -*-
# pylint: disable=protected-access
from threading import local
from functools import wraps
from dbquery.postgres import PostgresDB
from psycopg2.pool import ThreadedConnectionPool

import config


_CONNECTION_POOL = None


def _ensure_local(var_name, default):
    """ Ensure that var_name exists in _local on self, if not initialize it
    with default.
    """
    def _decorator(f):
        @wraps(f)
        def _new_f(self, *args, **kwds):
            if not hasattr(self._local, var_name):
                setattr(self._local, var_name, default)
            return f(self, *args, **kwds)
        return _new_f
    return _decorator


class _PostgresConnectionPoolDB(PostgresDB):

    def __init__(self, *args, **kwds):
        self._local = local()
        self._local._connection = None  # never use directly, use property!
        self._local._transaction_level = 0  # never use directly, use property!
        self._local._orig_retry = None  # never use directly, use property!
        super().__init__(*args, **kwds)

    @property
    @_ensure_local("_transaction_level", 0)
    def _transaction_level(self):
        return self._local._transaction_level

    @_transaction_level.setter
    @_ensure_local("_transaction_level", 0)
    def _transaction_level(self, value):
        self._local._transaction_level = value

    @property
    @_ensure_local("_connection", None)
    def _connection(self):
        return self._local._connection

    @_connection.setter
    @_ensure_local("_connection", None)
    def _connection(self, value):
        self._local._connection = value

    @property
    @_ensure_local("_orig_retry", None)
    def _orig_retry(self):
        return self._local._orig_retry

    @_orig_retry.setter
    @_ensure_local("_orig_retry", None)
    def _orig_retry(self, value):
        self._local._orig_retry = value

    def _connect(self):
        global _CONNECTION_POOL
        if _CONNECTION_POOL is None:
            _CONNECTION_POOL = ThreadedConnectionPool(
                config.DB_MIN_CONNECTIONS, config.DB_MAX_CONNECTIONS,
                **config.DB_PARAMS)
        if self._connection is not None:
            raise RuntimeError("Connection still exists.")
        self._connection = _CONNECTION_POOL.getconn()
        self._connection.set_session(autocommit=True)

    def close(self):
        if self._connection is not None:
            _CONNECTION_POOL.putconn(self._local._connection)
            self._connection = None
        super().close()


db = _PostgresConnectionPoolDB()



# -*- coding: utf-8 -*-
from logging import getLogger
from threading import Thread

from zmq import REQ, REP, Poller, POLLIN, Context, ROUTER, DEALER, proxy

import config
from translations_server.db import get_translation
from translations_server.lib import db


_LOG = getLogger(__name__)

_ENCODING = "utf-8"

_REQUEST_ENDPOINT = "inproc://requests"

_SYNC_ENDPOINT = "inproc://sync"


def _translate(lang, country, *key_plural_pairs):
    """ Generator: get translations from the DB. """
    # Group each two elements together to key and plural tuples.
    key_plural_pairs = zip(  # 1, 2, 3, 4, ... -> (1, 2), (3, 4), ...
        key_plural_pairs[::2], key_plural_pairs[1::2])
    for key, plural in key_plural_pairs:
        plural = int(plural) if plural else None
        translation = get_translation(lang, country, key, plural)
        if translation is None:
            extra = {
                "lang": lang,
                "country": country,
                "plural": plural
            }
            _LOG.warning("Missing translation: %s", key, extra=extra)
        yield translation or key


def _handle_request(parts):
    """ Handle one translation request, translate and return the result.

    In and out-puts are bytes.

    :param parts: Translation fields:
        - language
        - country or `""`
        - on or more pairs of key and plural form
    :type parts: [bytes]
    :return: The translations or `[b""]` if there was a encoding or format
        error.
    :rtype: [bytes] or [b""]
    """
    translations = [b""]  # error
    try:
        parts = [p.decode(_ENCODING) for p in parts]
    except UnicodeDecodeError:
        _LOG.exception("Decoding error", extra={"encoding": _ENCODING})
    else:
        if len(parts) < 4 or len(parts) % 2:
            _LOG.warning("Wrong count of arguments.")
        else:
            lang, country = parts[:2]
            try:
                translations = [
                    t.encode(_ENCODING)
                    for t in _translate(lang, country, *parts[2:])]
            except UnicodeEncodeError:
                _LOG.exception(
                    "Could not encode translation!",
                    extra={"encoding": _ENCODING})
            finally:
                # Close DB connection if any was opened.
                db.close()
    return translations


def _handle_requests(context):
    """ This is supposed to run as a background thread.

    It listens for translation requests and answers them until a message is
    arrived on the sync socket, whereupon the function (thread) ends.

    :type context: zmq.Context
    """
    sync_socket = context.socket(REQ)
    sync_socket.connect(_SYNC_ENDPOINT)
    requests_socket = context.socket(REP)
    requests_socket.connect(_REQUEST_ENDPOINT)

    _LOG.debug("Synchronizing worker")
    sync_socket.send(b"")
    sync_socket.recv()
    sync_socket.send(b"")

    keep_running = True
    poller = Poller()
    poller.register(requests_socket, POLLIN)
    poller.register(sync_socket, POLLIN)
    _LOG.debug("Running worker")
    while keep_running:
        sockets = dict(poller.poll())
        if requests_socket in sockets:
            try:
                response = _handle_request(requests_socket.recv_multipart())
            except Exception:  # pylint: disable=broad-except
                _LOG.critical("Handler crashed!", exc_info=True)
                response = [b""]
            requests_socket.send_multipart(response)
        if sync_socket in sockets:
            sync_socket.recv()
            keep_running = False
    _LOG.debug("Terminating worker")


def _start_workers(context, sync_socket, count, timeout=None):
    """
    :type context: zmq.Context
    :type sync_socker: zmq.Socket
    :type count: int
    :param timeout: Timeout for waiting for worker messages, in milliseconds.
    :type timeout: float
    """
    _LOG.debug("Starting workers...")
    worker_threads = [
        Thread(
            target=_handle_requests, name="worker {}".format(i),
            args=(context, ))
        for i in range(count)]
    for thread in worker_threads:
        thread.start()
    _LOG.debug("Synchronizing workers...")
    poller = Poller()
    poller.register(sync_socket, POLLIN)
    worker_identities = []
    for _ in worker_threads:
        sockets = dict(poller.poll(timeout=timeout))
        if sync_socket in sockets:
            worker_identities.append(sync_socket.recv_multipart()[0])
        else:
            raise RuntimeError("Worker did not respond in time.")
    for worker_identity in worker_identities:
        sync_socket.send_multipart([worker_identity, b"", b""])
    for _ in worker_identities:
        sockets = dict(poller.poll(timeout=timeout))
        if sync_socket in sockets:
            sync_socket.recv_multipart()
        else:
            raise RuntimeError("Worker did not respond in time.")
    _LOG.debug("Workers synchronized.")
    return worker_threads, worker_identities


def _shut_down_workers(
        sync_socket, worker_threads, worker_identities, timeout=None):
    """
    :type sync_socker: zmq.Socket
    :type worker_threads: [threading.Thread]
    :type worker_identities: [bytes]
    :param timeout: Timeout for waiting for worker threads, in seconds.
    :type timeout: float
    """
    _LOG.debug("Terminating workers...")
    for worker_identity in worker_identities:
        sync_socket.send_multipart([worker_identity, b"", b""])
    for i, thread in enumerate(worker_threads):
        thread.join(timeout=timeout)
        if thread.is_alive():
            raise RuntimeError("Worker {} did not terminate.".format(i))
    _LOG.debug("Workers terminated.")


def run(port):
    """ Run a translations server at a spesific port.

    It always listens on all available network devices!
    """
    context = Context(1)
    sync_socket = context.socket(ROUTER)
    sync_socket.bind(_SYNC_ENDPOINT)
    frontend = context.socket(ROUTER)
    frontend.bind("tcp://*:{}".format(port))
    # Socket facing services
    backend = context.socket(DEALER)
    backend.bind(_REQUEST_ENDPOINT)
    try:
        worker_threads, worker_identities = _start_workers(
            context, sync_socket, config.WORKERS, 1000)
        _LOG.debug("Running device...")
        try:
            proxy(frontend, backend)
        except KeyboardInterrupt:
            print("\rShutting down...")
        frontend.close()
        frontend = None
        _shut_down_workers(sync_socket, worker_threads, worker_identities, 5)
    finally:
        if frontend is not None:
            frontend.close()
        backend.close()
        sync_socket.close()
    _LOG.debug("Done")



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="TranslationsServer",
    version="0.1.1",
    description="The translations service server",
    long_description=long_description,
    url="https://github.com/GreenelyAB/TranslationsServer",
    author="GreenelyAB",
    author_email="info@greenely.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Communications",
        "Topic :: Software Development :: Internationalization",
        "Topic :: Software Development :: Localization",
        "Topic :: Text Processing",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
    ],
    keywords=(
        "store manage translation translations server internationalization"),
    packages=find_packages("src", exclude=["tests"]),
    package_dir = {"": "src"},
    install_requires=["pyzmq>=15.2.0", "psycopg2>=2.6.1", "DBQuery>=0.3.1"],
    extras_require={
    },
)
